module crypto_interface
open bytes

(* -------------- Programming Assumptions and TODO List -------------- *)
(* 0. Predicate ValidVote. Discuss about this on Friday. *)
(* 1. We consider IMPLICIT randomness and need(?) to implement Re-encryption *)
(* 2. We need to IMPLEMENT PUBLIC/SECRET KEYS and add them to predicate Encrypted *)
(* 3. We need an INTERFACE for cryptographic functions and IMPLEMENT them *)
(* 4. We have ONLY ONE vote for the attacker, we may give him more, and/or add Eq voters *)
(* 5. We have pretty IMPLICIT zk proofs, we may make them more explicit *)
(* 6. We assume ONLY TWO candidates for now. *)

(* -------------- Eq Types -------------- *)
type eq 'a = x:'a {| L x = R x |}

(* -------------- Specified constants -------------- *)
type vote = eq int   
val v1:vote
val v2:vote

(* -------------- Types Definitions -------------- *)
type cipher = eq bytes

(* -------------- Predicates Definitions -------------- *)
type Encrypted:: bytes => bytes => E
type Encryptedboth:: bytes => bytes => bytes => bytes => E
(* Note: We define a "both" version of predicates for writing purposes. *)

type FromA:: bytes => E
type FromAboth:: bytes => bytes => E

type FromB:: bytes => E
type FromBboth:: bytes => bytes => E

type FromO:: bytes => E
type FromOboth:: bytes => bytes => E

type Marsh:: int => bytes => E
type Marshboth:: int => int => bytes => bytes => E

type Valid:: bytes => E
type Validboth:: bytes => bytes => E

(* -------------- Macros -------------- *)
assume forall (m:bytes) (n:bytes) (x:bytes) (y:bytes).
    ( (Encryptedboth m n x y) <==> ( (Encrypted m x) && (Encrypted n y) ) )
assume forall (x:bytes) (y:bytes).( FromAboth x y <==> ( ( FromA x ) && ( FromA y ) ) )
assume forall (x:bytes) (y:bytes).( FromBboth x y <==> ( ( FromB x ) && ( FromB y ) ) )
assume forall (x:bytes) (y:bytes).( FromOboth x y <==> ( ( FromO x ) && ( FromO y ) ) )
assume forall (m:int) (n:int) (x:bytes) (y:bytes).
    ( (Marshboth m n x y) <==> ( (Marsh m x) && (Marsh n y) ) )
assume forall (x:bytes) (y:bytes).( Validboth x y <==> ( ( Valid x ) && ( Valid y ) ) )

(* -------------- Marshalling -------------- *)
(* Marshalling Assumptions *)
assume forall (z1:int) (z2:int) (m:bytes). ( ((Marsh z1 m) && (Marsh z2 m)) => (z1 = z2) )
assume forall (z:int) (m1:bytes) (m2:bytes). ( ((Marsh z m1) && (Marsh z m2)) => (m1 = m2) )

(* Marshalling Functions *)
val I2B: x:eq int -> y:bytes{| Marshboth (L x) (R x) (L y) (R y) |}
val B2I: x:bytes -> y:int{ Marsh y x }

(* -------------- Seal for randomized asymmetric encryption (simple decryption) -------------- *)
(* Seal reference for T(able) All *)
type seal_ref_entry_t =  m:bytes * c:cipher{| Encryptedboth (L m) (R m) (L c) (R c) |}
type seal_ref_list_t = list seal_ref_entry_t
type seal_ref_t = ref seal_ref_list_t

(* Seal reference for T(able) Valid *)
type seal_val_ref_entry_t =  m:bytes * c:cipher{| Encryptedboth (L m) (R m) (L c) (R c)
                && ( Validboth (L c) (R c) )
                && ( (FromAboth (L c) (R c)) || (FromBboth (L c) (R c)) 
					|| ( (FromOboth (L c) (R c)) && (L m = R m) ) )
				&& ( exists (zL:int) (zR:int).(Marshboth zL zR (L m) (R m)) )
                |}
(* Note: seal_fun_O_t will need to check that input is marsh of an int (v1 or v2). *)
                
type seal_val_ref_list_t = list seal_val_ref_entry_t
type seal_val_ref_t = ref seal_val_ref_list_t

(* Sealing functions types for T(able) all and valid *)
type seal_fun_A_t = m:bytes{| (Marsh (L v1) (L m)) && (Marsh (R v2) (R m)) |} -> c:cipher{|
                (Encryptedboth (L m) (R m) (L c) (R c)) && (FromAboth (L c) (R c)) 
                |}
(* does: new c.assume (FromAboth (L c) (R c)) && (Encrypted m c) && (Valid c) store (m,c) in both tables *)

val encA_in_progress: seal_fun_A_t

type seal_fun_B_t = m:bytes{| (Marsh (L v2) (L m)) && (Marsh (R v1) (R m)) |} -> c:cipher{|
                (Encryptedboth (L m) (R m) (L c) (R c)) && (FromBboth (L c) (R c))
                |}
				
val encB_in_progress: seal_fun_B_t

type seal_fun_O_t = m:bytes{| L m = R m |} -> cipher

val encO_in_progress: seal_fun_O_t

(* Unseal function type for T(able) all (will be used for decryption) *)
type unseal_fun_t =  c:cipher -> m:bytes{| forall (z1:bytes) (z2:bytes). (
                    (Encryptedboth z1 z2 (L c) (R c)) => ((z1 = L m) && (z2 = R m))
                    ) |}
(* Note: Should be an option type *)

val dec_in_progress : unseal_fun_t

(* Unseal function type for T(able) valid (used for validity checks) *)
type unseal_val_fun_t =  c:cipher -> b:bool{| (( (L b) = true ) && ( (R b) = true ))  => (
                        (Validboth (L c) (R c)) 
						&& (exists (m1:bytes) (m2:bytes) (z1:int) (z2:int).( 
							(Encryptedboth m1 m2 (L c) (R c)) && (Marshboth z1 z2 m1 m2) 
							&& ( ( (FromAboth (L c) (R c)) )
								|| ( (FromBboth (L c) (R c)) )
								|| ( (FromOboth (L c) (R c)) && (m1 = m2) ) )
						))
						)|}

val check_val_in_progress : unseal_val_fun_t

(* Reseal function type for homomorphic encryption *)
type reseal_hom_fun_t = c1:cipher -> c2:cipher -> c:cipher{|
                    forall (mL1:bytes) (mL2:bytes) (mR1:bytes) (mR2:bytes) 
							(zL1:int) (zL2:int) (zR1:int) (zR2:int).(
					( (Encryptedboth mL1 mR1 (L c1) (R c1)) && (Encryptedboth mL2 mR2 (L c2) (R c2))
					&& (Marshboth zL1 zR1 mL1 mR1) && (Marshboth zL2 zR2 mL2 mR2) )
                    => (exists (mL:bytes) (mR:bytes).(
						(Encryptedboth mL mR (L c) (R c)) && (Marshboth (zL1 + zL2) (zR1 + zR2) mL mR) 
						))
					)|}
                    
val homo_in_progress : reseal_hom_fun_t

(* -------------- Trash -------------- *)
(* COMMENT: old version not working (?). Seems same problem as Alice.fst *)
(*
type reseal_hom_fun_t = c1:cipher -> c2:cipher -> c:cipher{|
                    forall (mL1:bytes) (mL2:bytes) (mR1:bytes) (mR2:bytes) 
							(zL1:int) (zL2:int) (zR1:int) (zR2:int).(
					( (Encryptedboth mL1 mR1 (L c1) (R c1)) && (Encryptedboth mL2 mR2 (L c2) (R c2))
					&& (Marshboth zL1 zR1 mL1 mR1) && (Marshboth zL2 zR2 mL2 mR2) )
                    => (exists (mL:bytes) (mR:bytes) (zL:int) (zR:int).(
						(Encryptedboth mL mR (L c) (R c)) && (Marshboth zL zR mL mR) 
						&& (zL = (zL1 + zL2)) && (zR = (zR1 + zR2))
						))
					)|}
*)